from pathlib import Path
from re import compile, IGNORECASE

from snowddl.blueprint import TableBlueprint, TableColumn, PrimaryKeyBlueprint, UniqueKeyBlueprint, ForeignKeyBlueprint, DataType, Ident, SchemaObjectIdent, TableConstraintIdent, build_schema_object_ident
from snowddl.config import SnowDDLConfig
from snowddl.parser.abc_parser import AbstractParser, ParsedFile
from snowddl.parser.schema import database_json_schema, schema_json_schema

col_type_re = compile(r'^(?P<type>[a-z0-9_]+(\((\d+)(,(\d+))?\))?)'
                      r'(?P<not_null> NOT NULL)?$'
                      , IGNORECASE)


table_json_schema = {
    "type": "object",
    "properties": {
        "columns": {
            "type": "object",
            "minItems": 1,
            "items": {
                "anyOf": [
                    {
                        "type": "string"
                    },
                    {
                        "type": "object",
                        "properties": {
                            "type": {
                                "type": "string"
                            },
                            "default": {
                                "type": "string"
                            },
                            "default_sequence": {
                                "type": "string"
                            },
                            "collate": {
                                "type": "string"
                            },
                            "comment": {
                                "type": "string"
                            }
                        },
                        "required": ["type"],
                        "additionalProperties": False
                    }
                ]
            }
        },
        "cluster_by": {
            "type": "array",
            "items": {
                "type": "string"
            }
        },
        "change_tracking": {
            "type": "boolean"
        },
        "search_optimization": {
            "type": "boolean"
        },
        "comment": {
            "type": "string"
        },
        "primary_key": {
            "type": "array",
            "items": {
                "type": "string"
            },
            "minItems": 1
        },
        "unique_keys": {
            "type": "array",
            "items": {
                "type": "array",
                "items": {
                    "type": "string"
                },
                "minItems": 1,
            }
        },
        "foreign_keys": {
            "type": "array",
            "items": {
                "type": "object",
                "properties": {
                    "columns": {
                        "type": "array",
                        "items": {
                            "type": "string"
                        },
                        "minItems": 1
                    },
                    "ref_table": {
                        "type": "string"
                    },
                    "ref_columns": {
                        "type": "array",
                        "items": {
                            "type": "string"
                        },
                        "minItems": 1
                    }
                },
                "additionalProperties": False
            },
            "minItems": 1
        }
    },
    "additionalProperties": False
}


class TableParser(AbstractParser):
    def __init__(self, config: SnowDDLConfig, base_path: Path):
        super().__init__(config, base_path)
        self.combined_params = self.init_combined_params()

    def init_combined_params(self):
        combined_params = {}

        for database_path in self.base_path.iterdir():
            if not database_path.is_dir():
                continue

            database_params = self.parse_single_file(database_path / 'params.yaml', database_json_schema)
            combined_params[database_path.name] = {}

            for schema_path in database_path.iterdir():
                if not schema_path.is_dir():
                    continue

                schema_params = self.parse_single_file(schema_path / 'params.yaml', schema_json_schema)

                combined_params[database_path.name][schema_path.name] = {
                    "is_transient": database_params.get('is_transient', False) or schema_params.get('is_transient', False),
                    "retention_time": schema_params.get("retention_time"),
                }

        return combined_params

    def load_blueprints(self):
        self.parse_schema_object_files('table', table_json_schema, self.process_table)

    def process_table(self, f: ParsedFile):
        column_blueprints = []

        for col_name, col in f.params['columns'].items():

            # Short syntax
            if isinstance(col, str):
                col = {
                    "type": col
                }

            m = col_type_re.match(col['type'])

            if not m:
                raise ValueError(f"Incorrect short syntax for column [{col_name}] in table [{f.database}.{f.schema}.{f.name}]: {col['type']}")

            # Default for sequence
            if col.get('default_sequence'):
                col_default = build_schema_object_ident(self.env_prefix, col['default_sequence'], f.database, f.schema)
            else:
                col_default = col.get('default')

            column_blueprints.append(
                TableColumn(
                    name=Ident(col_name),
                    type=DataType(m.group('type')),
                    not_null=bool(m.group('not_null')),
                    default=col_default,
                    collate=col.get('collate').lower() if col.get('collate') else None,
                    comment=col.get('comment'),
                )
            )

        bp = TableBlueprint(
            full_name=SchemaObjectIdent(self.env_prefix, f.database, f.schema, f.name),
            columns=column_blueprints,
            cluster_by=f.params.get('cluster_by', None),
            is_transient=f.params.get('is_transient', self.combined_params[f.database][f.schema].get('is_transient', False)),
            retention_time=f.params.get('retention_time', self.combined_params[f.database][f.schema].get('retention_time', None)),
            change_tracking=f.params.get('change_tracking', False),
            search_optimization=f.params.get('search_optimization', False),
            comment=f.params.get('comment'),
        )

        self.config.add_blueprint(bp)

        # Constraints

        if f.params.get('primary_key'):
            bp = PrimaryKeyBlueprint(
                full_name=TableConstraintIdent(self.env_prefix, f.database, f.schema, f.name, columns=[Ident(c) for c in f.params.get('primary_key')]),
                table_name=SchemaObjectIdent(self.env_prefix, f.database, f.schema, f.name),
                columns=[Ident(c) for c in f.params.get('primary_key')],
                comment=None,
            )

            self.config.add_blueprint(bp)

        for columns in f.params.get('unique_keys', []):
            bp = UniqueKeyBlueprint(
                full_name=TableConstraintIdent(self.env_prefix, f.database, f.schema, f.name, columns=[Ident(c) for c in columns]),
                table_name=SchemaObjectIdent(self.env_prefix, f.database, f.schema, f.name),
                columns=[Ident(c) for c in columns],
                comment=None,
            )

            self.config.add_blueprint(bp)

        for fk in f.params.get('foreign_keys', []):
            bp = ForeignKeyBlueprint(
                full_name=TableConstraintIdent(self.env_prefix, f.database, f.schema, f.name, columns=[Ident(c) for c in fk['columns']]),
                table_name=SchemaObjectIdent(self.env_prefix, f.database, f.schema, f.name),
                columns=[Ident(c) for c in fk['columns']],
                ref_table_name=build_schema_object_ident(self.env_prefix, fk['ref_table'], f.database, f.schema),
                ref_columns=[Ident(c) for c in fk['ref_columns']],
                comment=None,
            )

            self.config.add_blueprint(bp)



from dataclasses import dataclass
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .ident import AbstractIdent
    from .object_type import ObjectType


@dataclass
class Grant:
    privilege: str
    on: "ObjectType"
    name: "AbstractIdent"


@dataclass
class FutureGrant:
    privilege: str
    on: "ObjectType"
    name: "AbstractIdent"



from snowddl.blueprint import TaskBlueprint, AccountObjectIdent, SchemaObjectIdent, build_schema_object_ident
from snowddl.parser.abc_parser import AbstractParser, ParsedFile


task_json_schema = {
    "type": "object",
    "properties": {
        "body": {
            "type": "string"
        },
        "schedule": {
            "type": "string"
        },
        "after": {
            "type": "string"
        },
        "when": {
            "type": "string"
        },
        "warehouse": {
            "type": "string"
        },
        "user_task_managed_initial_warehouse_size": {
            "type": "string"
        },
        "allow_overlapping_execution": {
            "type": "boolean"
        },
        "session_params": {
            "type": "object",
            "additionalProperties": {
                "type": ["boolean", "number", "string"]
            }
        },
        "user_task_timeout_ms": {
            "type": "integer"
        },
        "comment": {
            "type": "string"
        }
    },
    "additionalProperties": False
}


class TaskParser(AbstractParser):
    def load_blueprints(self):
        self.parse_schema_object_files("task", task_json_schema, self.process_task)

    def process_task(self, f: ParsedFile):
        if f.params.get('after'):
            task_after_ident = build_schema_object_ident(self.env_prefix, f.params['after'], f.database, f.schema)
        else:
            task_after_ident = None

        bp = TaskBlueprint(
            full_name=SchemaObjectIdent(self.env_prefix, f.database, f.schema, f.name),
            body=f.params['body'],
            schedule=f.params.get('schedule'),
            after=task_after_ident,
            depends_on=[task_after_ident] if task_after_ident else [],
            when=f.params.get('when'),
            warehouse=AccountObjectIdent(self.env_prefix, f.params['warehouse']) if f.params.get('warehouse') else None,
            user_task_managed_initial_warehouse_size=f.params.get('user_task_managed_initial_warehouse_size'),
            allow_overlapping_execution=f.params.get('allow_overlapping_execution'),
            session_params=self.normalise_params_dict(f.params.get('session_params')),
            user_task_timeout_ms=f.params.get('user_task_timeout_ms'),
            comment=f.params.get('comment'),
        )

        self.config.add_blueprint(bp)



from snowddl.blueprint import FileFormatBlueprint
from snowddl.resolver.abc_schema_object_resolver import AbstractSchemaObjectResolver, ResolveResult, ObjectType


class FileFormatResolver(AbstractSchemaObjectResolver):
    def get_object_type(self) -> ObjectType:
        return ObjectType.FILE_FORMAT

    def get_existing_objects_in_schema(self, schema: dict):
        existing_objects = {}

        cur = self.engine.execute_meta("SHOW FILE FORMATS IN SCHEMA {database:i}.{schema:i}", {
            "database": schema['database'],
            "schema": schema['schema'],
        })

        for r in cur:
            existing_objects[f"{r['database_name']}.{r['schema_name']}.{r['name']}"] = {
                "database": r['database_name'],
                "schema": r['schema_name'],
                "name": r['name'],
                "owner": r['owner'],
                "type": r['type'],
                "format_options": r['format_options'],
                "comment": r['comment'],
            }

        return existing_objects

    def get_blueprints(self):
        return self.config.get_blueprints_by_type(FileFormatBlueprint)

    def create_object(self, bp: FileFormatBlueprint):
        query = self._build_create_file_format(bp)

        self.engine.execute_safe_ddl(query)

        self.engine.execute_safe_ddl("COMMENT ON FILE FORMAT {full_name:i} IS {comment}", {
            "full_name": bp.full_name,
            "comment": query.add_short_hash(bp.comment),
        })

        return ResolveResult.CREATE

    def compare_object(self, bp: FileFormatBlueprint, row: dict):
        query = self._build_create_file_format(bp)

        if not query.compare_short_hash(row['comment']):
            self.engine.execute_safe_ddl(query)

            self.engine.execute_safe_ddl("COMMENT ON FILE FORMAT {full_name:i} IS {comment}", {
                "full_name": bp.full_name,
                "comment": query.add_short_hash(bp.comment),
            })

            return ResolveResult.ALTER

        return ResolveResult.NOCHANGE

    def drop_object(self, row: dict):
        self.engine.execute_safe_ddl("DROP FILE FORMAT {database:i}.{schema:i}.{name:i}", {
            "database": row['database'],
            "schema": row['schema'],
            "name": row['name'],
        })

        return ResolveResult.DROP

    def _build_create_file_format(self, bp: FileFormatBlueprint):
        query = self.engine.query_builder()

        query.append("CREATE FILE FORMAT {full_name:i}", {
            "full_name": bp.full_name,
        })

        query.append_nl("TYPE = {type}", {
            "type": bp.type,
        })

        for option_name, option_value in bp.format_options.items():
            query.append_nl("{option_name:r} = {option_value:dp}", {
                "option_name": option_name,
                "option_value": option_value,
            })

        if bp.comment:
            query.append_nl("COMMENT = {comment}", {
                "comment": bp.comment,
            })

        return query



from collections import defaultdict
from pathlib import Path
from typing import List, Dict, Type, Union

from snowddl.blueprint import AbstractBlueprint, T_Blueprint


class SnowDDLConfig:
    BUSINESS_ROLE_SUFFIX = 'B_ROLE'
    SCHEMA_ROLE_SUFFIX = 'S_ROLE'
    TECH_ROLE_SUFFIX = 'T_ROLE'
    USER_ROLE_SUFFIX = 'U_ROLE'
    WAREHOUSE_ROLE_SUFFIX = 'W_ROLE'

    def __init__(self, env_prefix=None):
        self.env_prefix = self._init_env_prefix(env_prefix)

        self.blueprints: Dict[Type[T_Blueprint], Dict[str,T_Blueprint]] = defaultdict(dict)
        self.errors: List[dict] = []

        self.placeholders: Dict[str,Union[bool,float,int,str]] = {}

    def get_blueprints_by_type(self, cls: Type[T_Blueprint]) -> Dict[str,T_Blueprint]:
        return self.blueprints.get(cls, {})

    def get_placeholder(self, name: str) -> Union[bool,float,int,str]:
        if name not in self.placeholders:
            raise ValueError(f"Unknown placeholder [{name}]")

        return self.placeholders[name]

    def add_blueprint(self, bp: AbstractBlueprint):
        self.blueprints[bp.__class__][str(bp.full_name)] = bp

    def add_error(self, path: Path, e: Exception):
        self.errors.append({
            "path": path,
            "error": e,
        })

    def add_placeholder(self, name: str, value: Union[bool,float,int,str]):
        self.placeholders[name] = value

    def _init_env_prefix(self, env_prefix):
        if env_prefix:
            env_prefix = str(env_prefix).upper()

            if '__' in env_prefix:
                raise ValueError(f"Env prefix [{env_prefix}] cannot contain [__] double underscore")

            if env_prefix.endswith('_'):
                raise ValueError(f"Env prefix [{env_prefix}] cannot end with [_] underscore")

            return f"{env_prefix}__"

        return ''



from snowddl.blueprint import ObjectType
from snowddl.converter.abc_schema_object_converter import AbstractSchemaObjectConverter, ConvertResult
from snowddl.parser.sequence import sequence_json_schema


class SequenceConverter(AbstractSchemaObjectConverter):
    def get_object_type(self) -> ObjectType:
        return ObjectType.SEQUENCE

    def get_existing_objects_in_schema(self, schema: dict):
        existing_objects = {}

        cur = self.engine.execute_meta("SHOW SEQUENCES IN SCHEMA {database:i}.{schema:i}", {
            "database": schema['database'],
            "schema": schema['schema'],
        })

        for r in cur:
            existing_objects[f"{r['database_name']}.{r['schema_name']}.{r['name']}"] = {
                "database": r['database_name'],
                "schema": r['schema_name'],
                "name": r['name'],
                "owner": r['owner'],
                "next_value": r['next_value'],
                "interval": r['interval'],
                "comment": r['comment'] if r['comment'] else None,
            }

        return existing_objects

    def dump_object(self, row):
        data = {
            "interval": row['interval'],
            "comment": row['comment'],
        }

        object_path = self.base_path / self._normalise_name_with_prefix(row['database']) / self._normalise_name(row['schema']) / 'sequence'
        object_path.mkdir(mode=0o755, parents=True, exist_ok=True)

        if data:
            self._dump_file(object_path / f"{self._normalise_name(row['name'])}.yaml", data, sequence_json_schema)
            return ConvertResult.DUMP

        return ConvertResult.EMPTY



from snowddl.blueprint import UserBlueprint, build_role_ident
from snowddl.error import SnowDDLExecuteError
from snowddl.resolver.abc_resolver import AbstractResolver, ResolveResult, ObjectType
from snowddl.resolver._utils import compare_dynamic_param_value


class UserResolver(AbstractResolver):
    def get_object_type(self) -> ObjectType:
        return ObjectType.USER

    def get_existing_objects(self):
        existing_objects = {}

        cur = self.engine.execute_meta("SHOW USERS LIKE {env_prefix:ls}", {
            'env_prefix': self.config.env_prefix,
        })

        for r in cur:
            if r['owner'] != self.engine.context.current_role:
                continue

            existing_objects[r['name']] = {
                "name": r['name'],
                "login_name": r['login_name'] if r['login_name'] else None,
                "display_name": r['display_name'] if r['display_name'] else None,
                "first_name": r['first_name'] if r['first_name'] else None,
                "last_name": r['last_name'] if r['last_name'] else None,
                "email": r['email'] if r['email'] else None,
                "disabled": r['disabled'] == 'true',
                "default_warehouse": r['default_warehouse'] if r['default_warehouse'] else None,
                "default_namespace": r['default_namespace'] if r['default_namespace'] else None,
                "default_role": r['default_role'] if r['default_role'] else None,
                "has_password": r['has_password'] == 'true',
                "has_rsa_public_key": r['has_rsa_public_key'] == 'true',
            }

        return existing_objects

    def get_blueprints(self):
        return self.config.get_blueprints_by_type(UserBlueprint)

    def create_object(self, bp: UserBlueprint):
        query = self.engine.query_builder()

        query.append("CREATE USER {name:i}", {
            "name": bp.full_name,
        })

        if bp.password:
            query.append_nl("PASSWORD = {password}", {
                "password": bp.password,
            })

        if bp.rsa_public_key:
            query.append_nl("RSA_PUBLIC_KEY = {rsa_public_key}", {
                "rsa_public_key": bp.rsa_public_key,
            })

        if bp.rsa_public_key_2:
            query.append_nl("RSA_PUBLIC_KEY_2 = {rsa_public_key_2}", {
                "rsa_public_key_2": bp.rsa_public_key_2,
            })

        query.append(self._build_common_properties(bp))
        query.append(self._build_common_parameters(bp))

        self.engine.execute_safe_ddl(query)

        self.engine.execute_safe_ddl("GRANT ROLE {user_role:i} TO USER {user_name:i}", {
            "user_name": bp.full_name,
            "user_role": self._get_user_role_ident(bp),
        })

        return ResolveResult.CREATE

    def compare_object(self, bp: UserBlueprint, row: dict):
        result = ResolveResult.NOCHANGE

        if self._compare_properties(bp, row):
            result = ResolveResult.ALTER

        if self._compare_public_keys(bp, row):
            result = ResolveResult.ALTER

        if self._compare_parameters(bp):
            result = ResolveResult.ALTER

        if self._check_user_role_grant(bp):
            result = ResolveResult.ALTER

        if self.engine.settings.refresh_user_passwords and self._refresh_password(bp, row):
            result = ResolveResult.ALTER

        return result

    def drop_object(self, row: dict):
        self.engine.execute_unsafe_ddl("DROP USER {name:i}", {
            "name": row['name'],
        })

        return ResolveResult.DROP

    def _build_common_properties(self, bp: UserBlueprint):
        query = self.engine.query_builder()

        query.append_nl("LOGIN_NAME = {login_name}", {"login_name": bp.login_name})
        query.append_nl("DISPLAY_NAME = {display_name}", {"display_name": bp.display_name})

        query.append_nl("DISABLED = {disabled:b}", {"disabled": bp.disabled})
        query.append_nl("FIRST_NAME = {first_name}", {"first_name": bp.first_name})
        query.append_nl("LAST_NAME = {last_name}", {"last_name": bp.last_name})
        query.append_nl("EMAIL = {email}", {"email": bp.email})

        query.append_nl("DEFAULT_WAREHOUSE = {default_warehouse}", {"default_warehouse": bp.default_warehouse})
        query.append_nl("DEFAULT_NAMESPACE = {default_namespace}", {"default_namespace": bp.default_namespace})

        query.append_nl("DEFAULT_ROLE = {default_role:i}", {"default_role": self._get_user_role_ident(bp)})

        return query

    def _build_common_parameters(self, bp: UserBlueprint):
        query = self.engine.query_builder()

        for param_name, param_value in bp.session_params.items():
            query.append_nl("{param_name:r} = {param_value:dp}", {
                "param_name": param_name,
                "param_value": param_value,
            })

        return query

    def _compare_properties(self, bp: UserBlueprint, row: dict):
        if bp.login_name == row['login_name'] \
        and bp.display_name == row['display_name'] \
        and bp.first_name == row['first_name'] \
        and bp.last_name == row['last_name'] \
        and bp.email == row['email'] \
        and bp.disabled == row['disabled'] \
        and ((bp.default_warehouse is None and row['default_warehouse'] is None) or str(bp.default_warehouse) == row['default_warehouse']) \
        and ((bp.default_namespace is None and row['default_namespace'] is None) or str(bp.default_namespace) == row['default_namespace']) \
        and str(self._get_user_role_ident(bp)) == row['default_role']:
            return False

        query = self.engine.query_builder()

        query.append("ALTER USER {name:i} SET", {
            "name": bp.full_name,
        })

        query.append(self._build_common_properties(bp))

        self.engine.execute_safe_ddl(query)

        return True

    def _compare_public_keys(self, bp: UserBlueprint, row: dict):
        if not bp.rsa_public_key and not bp.rsa_public_key_2 and not row['has_rsa_public_key']:
            return False

        existing_public_key = None
        existing_public_key_2 = None

        cur = self.engine.execute_meta("DESC USER {name:i}", {
            "name": bp.full_name,
        })

        for r in cur:
            if r['property'] == 'RSA_PUBLIC_KEY' and r['value'] != 'null':
                existing_public_key = r['value']

            if r['property'] == 'RSA_PUBLIC_KEY_2' and r['value'] != 'null':
                existing_public_key_2 = r['value']

        if bp.rsa_public_key == existing_public_key \
        and bp.rsa_public_key_2 == existing_public_key_2:
            return False

        self.engine.execute_safe_ddl("ALTER USER {name:i} SET rsa_public_key = {rsa_public_key}, rsa_public_key_2={rsa_public_key_2}", {
            "name": bp.full_name,
            "rsa_public_key": bp.rsa_public_key,
            "rsa_public_key_2": bp.rsa_public_key_2,
        })

        return True

    def _compare_parameters(self, bp: UserBlueprint):
        existing_params = self._get_existing_user_parameters(bp)
        query = self.engine.query_builder()

        query.append("ALTER USER {name:i} SET", {
            "name": bp.full_name,
        })

        for param_name, param_value in bp.session_params.items():
            if param_name not in existing_params:
                raise ValueError(f"Unknown parameter [{param_name}] for user [{bp.full_name}]")

            if compare_dynamic_param_value(param_value, existing_params[param_name]['value']):
                continue

            # At least one parameter in blueprint does not match the existing value
            # Refresh all parameters for such user
            query.append(self._build_common_parameters(bp))
            break

        # If parameter was set on USER level and does not exist in blueprint anymore
        # Unset such parameter and reset it to default
        for param_name, p in existing_params.items():
            if p['level'] == 'USER' and p['key'] not in bp.session_params:
                # Setting parameter to NULL equals to UNSET
                query.append_nl("{param_name:r} = NULL", {
                    "param_name": param_name,
                })

        if query.fragment_count() > 1:
            self.engine.execute_safe_ddl(query)
            return True

        return False

    def _check_user_role_grant(self, bp: UserBlueprint):
        user_role = self._get_user_role_ident(bp)

        cur = self.engine.execute_meta("SHOW GRANTS TO USER {name:i}", {
            "name": bp.full_name,
        })

        for r in cur:
            if r['role'] == str(user_role):
                return False

        self.engine.execute_safe_ddl("GRANT ROLE {user_role:i} TO USER {user_name:i}", {
            "user_name": bp.full_name,
            "user_role": user_role,
        })

        return True

    def _refresh_password(self, bp: UserBlueprint, row: dict):
        if bp.password:
            try:
                self.engine.execute_safe_ddl("ALTER USER {name:i} SET PASSWORD = {password}", {
                    "name": bp.full_name,
                    "password": bp.password,
                })
            except SnowDDLExecuteError as e:
                # Password rejected due to 'PRIOR_USE'
                # Not a error, skip such user
                if e.snow_exc.errno == 3002:
                    return False
                else:
                    raise

            return True

        if row['has_password']:
            self.engine.execute_safe_ddl("ALTER USER {name:i} UNSET PASSWORD", {
                "name": bp.full_name,
            })

            return True

        return False

    def _get_user_role_ident(self, bp: UserBlueprint):
        return build_role_ident(self.config.env_prefix, bp.full_name, self.config.USER_ROLE_SUFFIX)

    def _get_existing_user_parameters(self, bp: UserBlueprint):
        existing_params = {}

        cur = self.engine.execute_meta("SHOW PARAMETERS IN USER {name:i}", {
            "name": bp.full_name,
        })

        for r in cur:
            existing_params[r['key']] = r

        return existing_params



from snowddl.blueprint import TaskBlueprint
from snowddl.resolver.abc_schema_object_resolver import AbstractSchemaObjectResolver, ResolveResult, ObjectType


class TaskResolver(AbstractSchemaObjectResolver):
    def get_object_type(self) -> ObjectType:
        return ObjectType.TASK

    def get_existing_objects_in_schema(self, schema: dict):
        existing_objects = {}

        cur = self.engine.execute_meta("SHOW TASKS IN SCHEMA {database:i}.{schema:i}", {
            "database": schema['database'],
            "schema": schema['schema'],
        })

        for r in cur:
            existing_objects[f"{r['database_name']}.{r['schema_name']}.{r['name']}"] = {
                "database": r['database_name'],
                "schema": r['schema_name'],
                "name": r['name'],
                "owner": r['owner'],
                "predecessors": r['predecessors'],
                "comment": r['comment'] if r['comment'] else None,
            }

        return existing_objects

    def get_blueprints(self):
        return self.config.get_blueprints_by_type(TaskBlueprint)

    def create_object(self, bp: TaskBlueprint):
        query = self._build_create_task(bp)

        self.engine.execute_safe_ddl(query)

        self.engine.execute_safe_ddl("COMMENT ON TASK {full_name:i} IS {comment}", {
            "full_name": bp.full_name,
            "comment": query.add_short_hash(bp.comment),
        })

        return ResolveResult.CREATE

    def compare_object(self, bp: TaskBlueprint, row: dict):
        query = self._build_create_task(bp)

        if not query.compare_short_hash(row['comment']):
            self.engine.execute_safe_ddl(query)

            self.engine.execute_safe_ddl("COMMENT ON TASK {full_name:i} IS {comment}", {
                "full_name": bp.full_name,
                "comment": query.add_short_hash(bp.comment),
            })

            return ResolveResult.REPLACE

        # Set predecessor again if it was dropped
        if bp.after and not row['predecessors']:
            self.engine.execute_safe_ddl("ALTER TASK {full_name:i} ADD AFTER {after:i}", {
                "full_name": bp.full_name,
                "after": bp.after,
            })

            return ResolveResult.ALTER

        return ResolveResult.NOCHANGE

    def drop_object(self, row: dict):
        self.engine.execute_safe_ddl("DROP TASK {database:i}.{schema:i}.{name:i}", {
            "database": row['database'],
            "schema": row['schema'],
            "name": row['name'],
        })

        return ResolveResult.DROP

    def _build_create_task(self, bp: TaskBlueprint):
        query = self.engine.query_builder()

        query.append("CREATE OR REPLACE TASK {full_name:i}", {
            "full_name": bp.full_name,
        })

        if bp.warehouse:
            query.append_nl("WAREHOUSE = {warehouse:i}", {
                "warehouse": bp.warehouse,
            })

        if bp.user_task_managed_initial_warehouse_size:
            query.append_nl("USER_TASK_MANAGED_INITIAL_WAREHOUSE_SIZE = {size:r}", {
                "size": bp.user_task_managed_initial_warehouse_size,
            })

        if bp.schedule:
            query.append_nl("SCHEDULE = {schedule}", {
                "schedule": bp.schedule,
            })

        if bp.allow_overlapping_execution:
            query.append_nl("ALLOW_OVERLAPPING_EXECUTION = {allow_overlapping_execution:b}", {
                "allow_overlapping_execution": bp.allow_overlapping_execution,
            })

        if bp.user_task_timeout_ms:
            query.append_nl("USER_TASK_TIMEOUT_MS = {timeout:d}", {
                "timeout": bp.user_task_timeout_ms,
            })

        if bp.session_params:
            for param_name, param_value in bp.session_params.items():
                query.append_nl("{param_name:r} = {param_value:dp}", {
                    "param_name": param_name,
                    "param_value": param_value,
                })

        if bp.comment:
            query.append_nl("COMMENT = {comment}", {
                "comment": bp.comment,
            })

        if bp.after:
            query.append_nl("AFTER {after:i}", {
                "after": bp.after
            })

        if bp.when:
            query.append_nl("WHEN {when:r}", {
                "when": bp.when
            })

        # Not supported by Snowflake, but documented
        # query.append_nl("COPY GRANTS")

        query.append_nl("AS")
        query.append_nl(bp.body)

        return query



from snowddl.blueprint import MaterializedViewBlueprint, Edition
from snowddl.resolver.abc_schema_object_resolver import AbstractSchemaObjectResolver, ResolveResult, ObjectType


class MaterializedViewResolver(AbstractSchemaObjectResolver):
    skip_min_edition = Edition.ENTERPRISE

    def get_object_type(self) -> ObjectType:
        return ObjectType.MATERIALIZED_VIEW

    def get_existing_objects_in_schema(self, schema: dict):
        existing_objects = {}

        cur = self.engine.execute_meta("SHOW MATERIALIZED VIEWS IN SCHEMA {database:i}.{schema:i}", {
            "database": schema['database'],
            "schema": schema['schema'],
        })

        for r in cur:
            existing_objects[f"{r['database_name']}.{r['schema_name']}.{r['name']}"] = {
                "database": r['database_name'],
                "schema": r['schema_name'],
                "name": r['name'],
                "owner": r['owner'],
                "text": r['text'],
                "is_secure": r['is_secure'] == "true",
                "cluster_by": r['cluster_by'] if r['cluster_by'] else None,
                "invalid": r['invalid'] == "true",
                "invalid_reason": r['invalid_reason'],
            }

        return existing_objects

    def get_blueprints(self):
        return self.config.get_blueprints_by_type(MaterializedViewBlueprint)

    def create_object(self, bp: MaterializedViewBlueprint):
        self.engine.execute_safe_ddl(self._build_create_materialized_view(bp))

        return ResolveResult.CREATE

    def compare_object(self, bp: MaterializedViewBlueprint, row: dict):
        query = self._build_create_materialized_view(bp)

        # If view text is exactly the same, and view is still valid, do nothing
        if row['text'] == str(query) and not row['invalid']:
            return ResolveResult.NOCHANGE

        self.engine.execute_unsafe_ddl(query)

        return ResolveResult.REPLACE

    def drop_object(self, row: dict):
        self.engine.execute_unsafe_ddl("DROP MATERIALIZED VIEW {database:i}.{schema:i}.{name:i}", {
            "database": row['database'],
            "schema": row['schema'],
            "name": row['name'],
        })

        return ResolveResult.DROP

    def _build_create_materialized_view(self, bp: MaterializedViewBlueprint):
        query = self.engine.query_builder()

        query.append("CREATE OR REPLACE")

        if bp.is_secure:
            query.append("SECURE")

        query.append("MATERIALIZED VIEW {full_name:i}", {
            "full_name": bp.full_name,
        })

        query.append_nl("COPY GRANTS")

        if bp.columns:
            query.append_nl('(')

            for idx, c in enumerate(bp.columns):
                query.append_nl("    {comma:r}{col_name:i}", {
                    "comma": "  " if idx == 0 else ", ",
                    "col_name": c.name,
                })

                if c.comment:
                    query.append("COMMENT {col_comment}", {
                        "col_comment": c.comment,
                    })

            query.append_nl(')')

        if bp.cluster_by:
            query.append_nl("CLUSTER BY ({cluster_by:r})", {
                "cluster_by": bp.cluster_by,
            })

        query.append_nl("AS")
        query.append_nl(bp.text)

        return query



from snowddl.blueprint import BusinessRoleBlueprint
from snowddl.resolver.abc_role_resolver import AbstractRoleResolver


class BusinessRoleResolver(AbstractRoleResolver):
    def get_role_suffix(self):
        return self.config.BUSINESS_ROLE_SUFFIX

    def get_blueprints(self):
        return self.config.get_blueprints_by_type(BusinessRoleBlueprint)



from ._parsed_file import ParsedFile
from .account_params import AccountParameterParser
from .business_role import BusinessRoleParser
from .database import DatabaseParser
from .external_function import ExternalFunctionParser
from .external_table import ExternalTableParser
from .file_format import FileFormatParser
from .function import FunctionParser
from .materialized_view import MaterializedViewParser
from .masking_policy import MaskingPolicyParser
from .network_policy import NetworkPolicyParser
from .pipe import PipeParser
from .placeholder import PlaceholderParser
from .procedure import ProcedureParser
from .resource_monitor import ResourceMonitorParser
from .row_access_policy import RowAccessPolicyParser
from .schema import SchemaParser
from .sequence import SequenceParser
from .stage import StageParser
from .stream import StreamParser
from .table import TableParser
from .task import TaskParser
from .tech_role import TechRoleParser
from .user import UserParser
from .view import ViewParser
from .warehouse import WarehouseParser


default_parser_sequence = [
    AccountParameterParser,
    NetworkPolicyParser,
    ResourceMonitorParser,
    WarehouseParser,
    DatabaseParser,
    SchemaParser,
    FileFormatParser,
    StageParser,
    SequenceParser,
    FunctionParser,
    ExternalFunctionParser,
    ProcedureParser,
    TableParser,
    ExternalTableParser,
    StreamParser,
    MaterializedViewParser,
    ViewParser,
    PipeParser,
    TaskParser,
    MaskingPolicyParser,
    RowAccessPolicyParser,
    TechRoleParser,
    BusinessRoleParser,
    UserParser,
]


singledb_parser_sequence = [
    DatabaseParser,
    SchemaParser,
    FileFormatParser,
    StageParser,
    SequenceParser,
    FunctionParser,
    ExternalFunctionParser,
    ProcedureParser,
    TableParser,
    ExternalTableParser,
    StreamParser,
    MaterializedViewParser,
    ViewParser,
    PipeParser,
    TaskParser,
    MaskingPolicyParser,
    RowAccessPolicyParser,
]



from snowddl.blueprint import Grant, TechRoleBlueprint, ObjectType, build_role_ident, build_grant_name_ident_config
from snowddl.parser.abc_parser import AbstractParser, ParsedFile


tech_role_json_schema = {
    "type": "object",
    "additionalProperties": {
        "type": "object",
        "properties": {
            "grants": {
                "type": "object",
                "additionalProperties": {
                    "type": "array",
                    "items": {
                        "type": "string"
                    },
                    "minItems": 1,
                    "comment": {
                        "type": "string"
                    }
                }
            },
            "comment": {
                "type": "string"
            }
        },
        "required": ["grants"],
        "additionalProperties": False
    }
}


class TechRoleParser(AbstractParser):
    def load_blueprints(self):
        self.parse_single_file(self.base_path / 'tech_role.yaml', tech_role_json_schema, self.process_tech_role)

    def process_tech_role(self, f: ParsedFile):
        for tech_role_name, tech_role in f.params.items():
            tech_role_ident = build_role_ident(self.env_prefix, tech_role_name, self.config.TECH_ROLE_SUFFIX)

            grants = []

            for definition, object_name_list in tech_role['grants'].items():
                on, privileges = definition.upper().split(':')

                for p in privileges.split(','):
                    for object_name in object_name_list:
                        grants.append(Grant(
                            privilege=p,
                            on=ObjectType[on],
                            name=build_grant_name_ident_config(self.env_prefix, object_name, ObjectType[on]),
                        ))

            bp = TechRoleBlueprint(
                full_name=tech_role_ident,
                grants=grants,
                future_grants=[],
                comment=tech_role.get('comment'),
            )

            self.config.add_blueprint(bp)



from typing import Union

from .data_type import BaseDataType
from .ident import AccountObjectIdent, DatabaseIdent, SchemaIdent, SchemaObjectIdent, SchemaObjectIdentWithArgs
from .object_type import ObjectType


def build_schema_object_ident(env_prefix, object_name, context_database_name, context_schema_name) -> SchemaObjectIdent:
    # Function or procedure identifier with arguments
    if object_name.endswith(')'):
        object_name, data_types_str = object_name.rstrip(')').split('(')
        data_types = [BaseDataType[dt.strip(' ').upper()] for dt in data_types_str.split(',')] if data_types_str else []
    else:
        data_types = None

    parts = object_name.split('.')
    parts_len = len(parts)

    if parts_len > 3:
        raise ValueError(f"Too many delimiters in schema object identifier [{object_name}]")

    # Add context db_name to "schema_name.object_name" or "object_name" identifier
    if parts_len <= 2:
        parts.insert(0, context_database_name)

    # Add context schema_name to "object_name" identifier
    if parts_len == 1:
        parts.insert(1, context_schema_name)

    # Function or procedure identifier with arguments
    if isinstance(data_types, list):
        return SchemaObjectIdentWithArgs(env_prefix, parts[0], parts[1], parts[2], data_types=data_types)

    return SchemaObjectIdent(env_prefix, parts[0], parts[1], parts[2])


def build_role_ident(env_prefix, *args: Union[AccountObjectIdent, str]) -> AccountObjectIdent:
    return AccountObjectIdent(env_prefix, f"{'__'.join(str(a.name) if isinstance(a, AccountObjectIdent) else str(a) for a in args)}")


def build_grant_name_ident_config(env_prefix, grant_name, object_type: ObjectType):
    parts = grant_name.split('.')
    last_part = parts[-1]

    if len(parts) == 3:
        # Extract data types for arguments of functions and procedures
        if '(' in last_part:
            start_dtypes_idx = last_part.index('(')
            finish_dtypes_idx = last_part.index(')')

            parts[-1] = last_part[0:start_dtypes_idx]

            data_types_str = last_part[start_dtypes_idx+1:finish_dtypes_idx]
            data_types = [BaseDataType[arg.strip(' ').upper()] for arg in data_types_str.split(',')] if data_types_str else []

            return SchemaObjectIdentWithArgs(env_prefix, parts[0], parts[1], parts[2], data_types=data_types)

        return SchemaObjectIdent(env_prefix, parts[0], parts[1], parts[2])

    if len(parts) == 2:
        return SchemaIdent(env_prefix, parts[0], parts[1])

    if len(parts) == 1:
        if object_type == ObjectType.DATABASE:
            return DatabaseIdent(env_prefix, parts[0])

        return AccountObjectIdent(env_prefix, parts[0])

    raise ValueError(f"Unexpected grant name format [{grant_name}] in config for object type [{object_type}]")


def build_grant_name_ident_snowflake(grant_name, object_type: ObjectType):
    env_prefix = ''

    parts = [p.strip('"') for p in grant_name.split('.')]
    last_part = parts[-1]

    # Remove object type component from future grant names
    if last_part.startswith('<') and last_part.endswith('>'):
        parts.pop()

    if len(parts) == 3:
        # Extract data types for arguments of functions and procedures
        if '(' in last_part:
            start_dtypes_idx = last_part.index('(')
            finish_dtypes_idx = last_part.index(')')

            parts[-1] = last_part[0:start_dtypes_idx]

            data_types_str = last_part[start_dtypes_idx+1:finish_dtypes_idx]
            data_types = [BaseDataType[arg.strip(' ').split(' ')[1]] for arg in data_types_str.split(',')] if data_types_str else []

            return SchemaObjectIdentWithArgs(env_prefix, parts[0], parts[1], parts[2], data_types=data_types)

        return SchemaObjectIdent(env_prefix, parts[0], parts[1], parts[2])

    if len(parts) == 2:
        return SchemaIdent(env_prefix, parts[0], parts[1])

    if len(parts) == 1:
        if object_type == ObjectType.DATABASE:
            return DatabaseIdent(env_prefix, parts[0])

        return AccountObjectIdent(env_prefix, parts[0])

    raise ValueError(f"Unexpected grant name format [{grant_name}] in Snowflake for object type [{object_type}]")


def build_default_namespace_ident(env_prefix, default_namespace):
    if '.' in default_namespace:
        return SchemaIdent(env_prefix, *default_namespace.split('.', 1))

    return DatabaseIdent(env_prefix, default_namespace)



from abc import ABC
from dataclasses import dataclass
from typing import Optional, List, Dict, Union, TypeVar, TYPE_CHECKING

from .column import ExternalTableColumn, TableColumn, ViewColumn, NameWithType
from .data_type import DataType
from .grant import Grant, FutureGrant
from .ident import AbstractIdent, Ident, AccountObjectIdent, DatabaseIdent, SchemaIdent, SchemaObjectIdent, SchemaObjectIdentWithArgs, StageFileIdent, TableConstraintIdent
from .reference import MaskingPolicyReference, RowAccessPolicyReference, TagReference
from .stage import StageWithPath

if TYPE_CHECKING:
    from .object_type import ObjectType


@dataclass
class AbstractBlueprint(ABC):
    full_name: AbstractIdent
    comment: Optional[str]


@dataclass
class SchemaObjectBlueprint(AbstractBlueprint, ABC):
    full_name: SchemaObjectIdent


@dataclass
class RoleBlueprint(AbstractBlueprint):
    full_name: AccountObjectIdent
    grants: List[Grant]
    future_grants: List[FutureGrant]


@dataclass
class DependsOnMixin(ABC):
    full_name: AbstractIdent
    depends_on: List[AbstractIdent]


@dataclass
class AccountParameterBlueprint(AbstractBlueprint):
    full_name: Ident
    value: Union[bool,float,int,str]


@dataclass
class BusinessRoleBlueprint(RoleBlueprint):
    pass


@dataclass
class DatabaseBlueprint(AbstractBlueprint):
    full_name: DatabaseIdent
    is_transient: Optional[bool]
    retention_time: Optional[int]
    is_sandbox: Optional[bool]


@dataclass
class ExternalFunctionBlueprint(SchemaObjectBlueprint):
    full_name: SchemaObjectIdent
    arguments: List[NameWithType]
    returns: DataType
    api_integraton: Ident
    url: str
    is_secure: bool
    is_strict: bool
    is_immutable: bool
    headers: Optional[Dict[str,str]]
    context_headers: Optional[List[Ident]]
    max_batch_rows: Optional[int]
    compression: Optional[str]
    request_translator: Optional[SchemaObjectIdent]
    response_translator: Optional[SchemaObjectIdent]


@dataclass
class ExternalTableBlueprint(SchemaObjectBlueprint):
    columns: Optional[List[ExternalTableColumn]]
    partition_by: Optional[List[Ident]]
    partition_type: Optional[str]
    location_stage: SchemaObjectIdent
    location_path: Optional[str]
    location_pattern: Optional[str]
    file_format: SchemaObjectIdent
    refresh_on_create: Optional[bool]
    auto_refresh: Optional[bool]
    aws_sns_topic: Optional[str]
    table_format: Optional[str]
    integration: Optional[Ident]


@dataclass
class FileFormatBlueprint(SchemaObjectBlueprint):
    type: str
    format_options: Optional[Dict[str,Union[bool,float,int,str,list]]]


@dataclass
class ForeignKeyBlueprint(SchemaObjectBlueprint):
    full_name: TableConstraintIdent
    table_name: SchemaObjectIdent
    columns: List[Ident]
    ref_table_name: SchemaObjectIdent
    ref_columns: List[Ident]


@dataclass
class FunctionBlueprint(SchemaObjectBlueprint):
    full_name: SchemaObjectIdentWithArgs
    language: str
    body: Optional[str]
    arguments: List[NameWithType]
    returns: Union[DataType,List[NameWithType]]
    is_secure: bool
    is_strict: bool
    is_immutable: bool
    runtime_version: Optional[str]
    imports: Optional[List[StageWithPath]]
    packages: Optional[List[str]]
    handler: Optional[str]


@dataclass
class MaterializedViewBlueprint(SchemaObjectBlueprint):
    text: str
    columns: Optional[List[ViewColumn]]
    is_secure: Optional[bool]
    cluster_by: Optional[List[str]]


@dataclass
class MaskingPolicyBlueprint(SchemaObjectBlueprint):
    full_name: SchemaObjectIdent
    arguments: List[NameWithType]
    returns: DataType
    body: str
    references: List[MaskingPolicyReference]


@dataclass
class NetworkPolicyBlueprint(AbstractBlueprint):
    full_name: Ident
    allowed_ip_list: List[str]
    blocked_ip_list: List[str]


@dataclass
class PipeBlueprint(SchemaObjectBlueprint):
    auto_ingest: bool
    copy_table_name: SchemaObjectIdent
    copy_stage_name: SchemaObjectIdent
    copy_path: Optional[str]
    copy_pattern: Optional[str]
    copy_transform: Optional[Dict[str, str]]
    copy_file_format: Optional[SchemaObjectIdent]
    copy_options: Optional[Dict[str,Union[bool,float,int,str,list]]]
    aws_sns_topic: Optional[str]
    integration: Optional[Ident]


@dataclass
class PrimaryKeyBlueprint(SchemaObjectBlueprint):
    full_name: TableConstraintIdent
    table_name: SchemaObjectIdent
    columns: List[Ident]


@dataclass
class ProcedureBlueprint(SchemaObjectBlueprint):
    full_name: SchemaObjectIdentWithArgs
    language: str
    body: str
    arguments: List[NameWithType]
    returns: Union[DataType,List[NameWithType]]
    is_strict: bool
    is_immutable: bool
    is_execute_as_caller: bool
    runtime_version: Optional[str]
    imports: Optional[List[StageWithPath]]
    packages: Optional[List[str]]
    handler: Optional[str]


@dataclass
class ResourceMonitorBlueprint(AbstractBlueprint):
    full_name: Ident
    credit_quota: int
    frequency: str
    triggers: Dict[int,str]


@dataclass
class RowAccessPolicyBlueprint(SchemaObjectBlueprint):
    full_name: SchemaObjectIdent
    arguments: List[NameWithType]
    body: str
    references: List[RowAccessPolicyReference]


@dataclass
class SchemaBlueprint(AbstractBlueprint):
    full_name: SchemaIdent
    is_transient: Optional[bool]
    retention_time: Optional[int]
    is_sandbox: Optional[bool]
    owner_additional_grants: List[Grant]


@dataclass
class SchemaRoleBlueprint(RoleBlueprint, DependsOnMixin):
    pass


@dataclass
class StageBlueprint(SchemaObjectBlueprint):
    url: Optional[str]
    storage_integration: Optional[Ident]
    encryption: Optional[Dict[str,str]]
    directory: Optional[Dict[str,str]]
    file_format: Optional[SchemaObjectIdent]
    copy_options: Optional[Dict[str,Union[bool,float,int,str,list]]]
    upload_stage_files: bool


@dataclass
class StageFileBlueprint(SchemaObjectBlueprint):
    full_name: StageFileIdent
    local_path: str
    stage_name: SchemaObjectIdent
    stage_path: str


@dataclass
class SequenceBlueprint(SchemaObjectBlueprint):
    start: int
    interval: int


@dataclass
class StreamBlueprint(SchemaObjectBlueprint):
    object_type: "ObjectType"
    object_name: SchemaObjectIdent
    append_only: Optional[bool]
    insert_only: Optional[bool]
    show_initial_rows: Optional[bool]


@dataclass
class TableBlueprint(SchemaObjectBlueprint):
    columns: List[TableColumn]
    cluster_by: Optional[List[str]]
    is_transient: bool
    retention_time: Optional[int]
    change_tracking: bool
    search_optimization: bool


@dataclass
class TagBlueprint(SchemaObjectBlueprint):
    full_name: SchemaObjectIdent
    references: List[TagReference]


@dataclass
class TaskBlueprint(SchemaObjectBlueprint, DependsOnMixin):
    body: str
    schedule: Optional[str]
    after: Optional[Ident]
    when: Optional[str]
    warehouse: Optional[AccountObjectIdent]
    user_task_managed_initial_warehouse_size: Optional[str]
    allow_overlapping_execution: Optional[bool]
    session_params: Optional[Dict[str,Union[bool,float,int,str]]]
    user_task_timeout_ms: Optional[int]


@dataclass
class TechRoleBlueprint(RoleBlueprint):
    pass


@dataclass
class UniqueKeyBlueprint(SchemaObjectBlueprint):
    full_name: TableConstraintIdent
    table_name: SchemaObjectIdent
    columns: List[Ident]


@dataclass
class UserBlueprint(AbstractBlueprint):
    full_name: AccountObjectIdent
    login_name: str
    display_name: str
    first_name: Optional[str]
    last_name: Optional[str]
    email: Optional[str]
    disabled: bool
    business_roles: List[AccountObjectIdent]
    password: Optional[str]
    rsa_public_key: Optional[str]
    rsa_public_key_2: Optional[str]
    default_warehouse: Optional[AccountObjectIdent]
    default_namespace: Optional[Union[DatabaseIdent, SchemaIdent]]
    session_params: Optional[Dict[str,Union[bool,float,int,str]]]


@dataclass
class ViewBlueprint(SchemaObjectBlueprint, DependsOnMixin):
    text: str
    columns: Optional[List[ViewColumn]]
    is_secure: Optional[bool]


@dataclass
class WarehouseBlueprint(AbstractBlueprint):
    full_name: AccountObjectIdent
    size: str
    auto_suspend: int
    min_cluster_count: Optional[int]
    max_cluster_count: Optional[int]
    scaling_policy: Optional[str]
    resource_monitor: Optional[Ident]


T_Blueprint = TypeVar('T_Blueprint', bound=AbstractBlueprint)



from snowddl.blueprint import TechRoleBlueprint
from snowddl.resolver.abc_role_resolver import AbstractRoleResolver


class TechRoleResolver(AbstractRoleResolver):
    def get_role_suffix(self):
        return self.config.TECH_ROLE_SUFFIX

    def get_blueprints(self):
        return self.config.get_blueprints_by_type(TechRoleBlueprint)



from abc import ABC, abstractmethod
from string import ascii_letters, digits
from typing import List, Optional, Tuple

from .data_type import BaseDataType


class AbstractIdent(ABC):
    allowed_chars = set(ascii_letters + digits + '_')

    @abstractmethod
    def __init__(self):
        pass

    @abstractmethod
    def parts_for_format(self) -> Tuple[List[str], Optional[List[str]]]:
        pass

    def __str__(self):
        core_parts, argument_parts = self.parts_for_format()

        if argument_parts is not None:
            return f"{'.'.join(core_parts)}({','.join(argument_parts)})"

        return '.'.join(core_parts)

    def __repr__(self):
        return f"<{self.__class__.__name__}={str(self)}>"

    def __eq__(self, other):
        if not isinstance(other, AbstractIdent):
            raise NotImplementedError

        return str(self) == str(other)

    def _validate_part(self, val):
        val = str(val)

        if not val:
            raise ValueError("Identifier cannot be empty")

        for char in val:
            if char not in self.allowed_chars:
                raise ValueError(f"Character [{char}] in not allowed in identifier [{val}], only ASCII letters, digits and single underscores are accepted")

        return val.upper()


class AbstractIdentWithPrefix(AbstractIdent, ABC):
    def __init__(self, env_prefix):
        self.env_prefix = self._validate_env_prefix(env_prefix)

    def _validate_env_prefix(self, val):
        val = str(val)

        for char in val:
            if char not in self.allowed_chars:
                raise ValueError(f"Character [{char}] in not allowed in env prefix [{val}], only ASCII letters, digits and single underscores are accepted")

        if val and not val.endswith('__'):
            raise ValueError(f"Env prefix [{val}] in identifier must end with [__] double underscore")

        return val.upper()


class Ident(AbstractIdent):
    def __init__(self, name):
        self.name = self._validate_part(name)

    def parts_for_format(self):
        return [self.name], None


class AccountObjectIdent(AbstractIdentWithPrefix):
    def __init__(self, env_prefix, name):
        super().__init__(env_prefix)

        self.name = self._validate_part(name)

    def parts_for_format(self):
        return [f"{self.env_prefix}{self.name}"], None


class DatabaseIdent(AbstractIdentWithPrefix):
    def __init__(self, env_prefix, database):
        super().__init__(env_prefix)

        self.database = self._validate_part(database)

    def parts_for_format(self):
        return [f"{self.env_prefix}{self.database}"], None


class SchemaIdent(AbstractIdentWithPrefix):
    def __init__(self, env_prefix, database, schema):
        super().__init__(env_prefix)

        self.database = self._validate_part(database)
        self.schema = self._validate_part(schema)

    def parts_for_format(self):
        return [f"{self.env_prefix}{self.database}", self.schema], None

    @property
    def database_full_name(self):
        return DatabaseIdent(self.env_prefix, self.database)


class SchemaObjectIdent(AbstractIdentWithPrefix):
    def __init__(self, env_prefix, database, schema, name):
        super().__init__(env_prefix)

        self.database = self._validate_part(database)
        self.schema = self._validate_part(schema)
        self.name = self._validate_part(name)

    def parts_for_format(self):
        return [f"{self.env_prefix}{self.database}", self.schema, self.name], None

    @property
    def database_full_name(self):
        return DatabaseIdent(self.env_prefix, self.database)


class SchemaObjectIdentWithArgs(SchemaObjectIdent):
    def __init__(self, env_prefix, database, schema, name, data_types: List[BaseDataType]):
        super().__init__(env_prefix, database, schema, name)

        self.data_types = data_types

    def parts_for_format(self):
        return [f"{self.env_prefix}{self.database}", self.schema, self.name], [data_type.name for data_type in self.data_types]


class StageFileIdent(SchemaObjectIdent):
    def __init__(self, env_prefix, database, schema, name, path):
        super().__init__(env_prefix, database, schema, name)

        self.path = path

    def parts_for_format(self):
        return [f"{self.env_prefix}{self.database}", self.schema, self.name], [self.path]


class TableConstraintIdent(SchemaObjectIdent):
    def __init__(self, env_prefix, database, schema, name, columns: List[Ident]):
        super().__init__(env_prefix, database, schema, name)

        self.columns = columns

    def parts_for_format(self):
        return [f"{self.env_prefix}{self.database}", self.schema, self.name], [str(c) for c in self.columns]
